'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.hashArgs = hashArgs;

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }

// attempt at making a standalone testable data structure

var _require = require('./helpers'),
    maybeGetSimpleString = _require.maybeGetSimpleString,
    getSimpleFragmentName = _require.getSimpleFragmentName;

// we could make razordata and bladedata inherit from a common class
// but honestly didnt want to prematurely optimize

/* eslint-disable complexity */
/* eslint-disable prefer-const */
/* eslint-disable no-use-before-define */

var RazorData = exports.RazorData = function () {
  function RazorData(_ref) {
    var _ref$args = _ref.args,
        args = _ref$args === undefined ? null : _ref$args,
        _ref$name = _ref.name,
        name = _ref$name === undefined ? null : _ref$name,
        _ref$type = _ref.type,
        type = _ref$type === undefined ? null : _ref$type,
        _ref$fragmentType = _ref.fragmentType,
        fragmentType = _ref$fragmentType === undefined ? null : _ref$fragmentType;

    _classCallCheck(this, RazorData);

    if (!type) throw new Error('type must be either fragment or query');
    if (type === 'fragment' && !fragmentType) throw new Error('fragments must come with a fragmentType');
    if (type === 'fragment' && !name) throw new Error('fragments must come with a name');
    this._children = []; // all the blades
    this._args = args; // a string for now
    this._name = name; // truly optional
    this._type = type; // either 'fragment' or 'query'
    this._fragmentType = fragmentType; // if fragment
  }

  RazorData.prototype.isFragment = function isFragment() {
    return this._type === 'fragment';
  };

  RazorData.prototype.getFragmentData = function getFragmentData() {
    return {
      name: this._name,
      fragmentType: this._fragmentType
    };
  };

  RazorData.prototype.get = function get(id) {
    for (var i = 0; i < this._children.length; i++) {
      var name = this._children[i]._name === id;
      var alias = this._children[i]._alias === id;
      if (name || alias) return this._children[i];
    }
    return null;
  };

  RazorData.prototype.add = function add(val) {
    var preferredNameOrAlias = val.args && val.args.length ? hashArgs(val.args, val.name) : val.name;
    var child = this.get(preferredNameOrAlias);
    // eslint-disable-next-line
    if (!(child && child._alias == hashArgs(val.args, val.name))) {
      // console.log('adding new child ', val.name, 'because', child && child._alias, 'vs', hashArgs(val.args, val.name))
      child = new BladeData(val);
      this._children.push(child);
    }
    return child; // return child so that more operations can be made
  };

  RazorData.prototype.print = function print() {
    var fields = this._children;
    if (!fields.length) return (
      /* eslint-disable-next-line */
      console.log('babel-blade Warning: razor with no children, doublecheck') || null
    ); // really shouldnt happen, should we throw an error?
    var maybeArgs = coerceNullLiteralToNull(this._args && this._args[0]);
    var TemplateLiteral = appendLiterals();
    if (this._type === 'query') {
      TemplateLiteral.addStr(`\nquery ${this._name || ''}`);
    } else {
      // have to make fragment name parametric
      TemplateLiteral.addStr(`\nfragment `);
      TemplateLiteral.addLit(this._name);
      TemplateLiteral.addStr(` on ${this._fragmentType}`);
    }
    TemplateLiteral.addStr(maybeArgs ? '(' : '').addLit(maybeArgs).addStr(maybeArgs ? ')' : '').addStr('{\n');

    var fragments = []; // will be mutated to add all the fragments included
    var accumulators = Object.keys(fields).map(function (key) {
      return fields[key].print('  ', fragments);
    });
    accumulators.forEach(TemplateLiteral.append);
    TemplateLiteral.addStr('}'); // cap off the string
    if (fragments.length) {
      fragments.forEach(function (frag) {
        TemplateLiteral.addStr('\n\n');
        TemplateLiteral.addLit(frag);
      });
    }
    return zipAccumulators(TemplateLiteral.get());
  };

  return RazorData;
}();

var BladeData = exports.BladeData = function () {
  function BladeData(_ref2) {
    var _ref2$name = _ref2.name,
        name = _ref2$name === undefined ? null : _ref2$name,
        _ref2$args = _ref2.args,
        args = _ref2$args === undefined ? [] : _ref2$args,
        _ref2$fragments = _ref2.fragments,
        fragments = _ref2$fragments === undefined ? [] : _ref2$fragments,
        _ref2$directives = _ref2.directives,
        directives = _ref2$directives === undefined ? [] : _ref2$directives;

    _classCallCheck(this, BladeData);

    if (!name) throw new Error('new Blade must have name');
    if (!Array.isArray(fragments)) throw new Error('fragments must be array');
    this._children = []; // store of child blades
    this._name = name; // a string for now
    this._args = args; // array
    this._alias = hashArgs(this._args, this._name);
    this._fragments = fragments.map(function (frag) {
      frag.isFragment = true;
      return frag;
    }); // tagging the literal as fragment for printing
    this._directives = directives;
  }

  BladeData.prototype.get = function get(id) {
    for (var i = 0; i < this._children.length; i++) {
      if (this._children[i]._name === id) return this._children[i];
      if (this._children[i]._alias === id) return this._children[i];
    }
    return null;
  };

  BladeData.prototype.add = function add(val) {
    var child = this.get(val.name);

    /* eslint-disable-next-line */
    if (!(child && child._alias == hashArgs(val.args, val.name))) {
      // console.log('adding new child2 because', child && child._alias, val.alias)
      child = new BladeData(val);
      this._children.push(child);
    }
    return child;
  };
  // TODO: potential problem here if blade has args/alias but no children


  BladeData.prototype.print = function print(indent, fragments) {
    var maybeArgs = this._args.length && this._args;
    var maybeDirs = this._directives.length && this._directives;
    var alias = this._alias;
    var printName = alias ? `${alias}: ${this._name}` : this._name;
    if (this._fragments.length) this._fragments.map(function (frag) {
      return fragments.push(frag);
    }); // mutates fragments!
    var TemplateLiteral = appendLiterals().addStr(`${indent}${printName}`).addStr(maybeArgs ? '(' : '');
    if (maybeArgs) {
      maybeArgs.forEach(function (arg, i) {
        if (i !== 0) TemplateLiteral.addStr(', ');
        TemplateLiteral.addLit(arg);
      });
    }
    TemplateLiteral.addStr(maybeArgs ? ')' : '');
    if (maybeDirs) {
      TemplateLiteral.addStr(' ');
      maybeDirs.forEach(function (dir, i) {
        if (i !== 0) TemplateLiteral.addStr(' ');
        TemplateLiteral.addLit(dir);
      });
    }
    var fields = this._children;
    if (fields.length || this._fragments.length) {
      TemplateLiteral.addStr(' {\n');
      var accumulators = Object.keys(fields).map(function (key) {
        return (
          /* eslint-disable-next-line */
          fields[key].print(indent + '  ', fragments)
        );
      });
      accumulators.forEach(TemplateLiteral.append);
      this._fragments.forEach(function (frag) {
        TemplateLiteral.addStr(`${indent}  ...${getSimpleFragmentName(frag)}\n`);
      });
      TemplateLiteral.addStr(`${indent}}\n`); // cap off the query
    } else {
      TemplateLiteral.addStr('\n');
    }
    return TemplateLiteral.get();
  };

  return BladeData;
}();

function hashArgs() {
  var args = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : [];
  var name = arguments[1];

  return args.length ? `${name}_${hashCode(JSON.stringify(args))}` : null;
}

// https://stackoverflow.com/a/8831937/1106414
function hashCode(str) {
  var hash = 0;
  if (str.length === 0) {
    return hash;
  }
  for (var i = 0; i < str.length; i++) {
    var char = str.charCodeAt(i);

    /* eslint-disable-next-line */
    hash = (hash << 5) - hash + char;

    /* eslint-disable-next-line */
    hash = hash & hash; // Convert to 32bit integer
  }
  return hash.toString(16).slice(-4); // last4hex
}

function appendLiterals() {
  var stringAccumulator = [];
  var litAccumulator = [];
  var me = {
    addStr() {
      var str = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : null;

      stringAccumulator.push(str);
      litAccumulator.push(null);
      return me;
    },
    addLit() {
      var lit = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : null;

      stringAccumulator.push(null);
      litAccumulator.push(lit);
      return me;
    },
    add() {
      var str = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : null;
      var lit = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : null;

      stringAccumulator.push(str);
      litAccumulator.push(lit);
      return me;
    },
    append(newMe) {
      newMe.stringAccumulator.forEach(function (str) {
        return stringAccumulator.push(str);
      });
      newMe.litAccumulator.forEach(function (lit) {
        return litAccumulator.push(lit);
      });
      return me;
    },
    get() {
      return { stringAccumulator, litAccumulator };
    }
  };
  return me;
}

function zipAccumulators(_ref3) {
  var stringAccumulator = _ref3.stringAccumulator,
      litAccumulator = _ref3.litAccumulator;

  // cannot have any spare

  /* eslint-disable-next-line */
  var str = '',
      newStrAcc = [],
      newLitAcc = [];
  for (var i = 0; i < stringAccumulator.length; i++) {
    if (litAccumulator[i]) {
      var maybeSimpleString = maybeGetSimpleString(litAccumulator[i]);
      if (maybeSimpleString) {
        // its just a simplestring!
        str += maybeSimpleString;
      } else {
        newLitAcc.push(litAccumulator[i]);
        newStrAcc.push(str + (stringAccumulator[i] || ''));
        str = '';
      }
    } else {
      // there is an empty lit, store in state
      str += stringAccumulator[i] || '';
    }
  }
  // flush store
  if (str !== '') newStrAcc.push(str);
  return { stringAccumulator: newStrAcc, litAccumulator: newLitAcc };
}

function coerceNullLiteralToNull(lit) {
  return lit && lit.type === 'NullLiteral' ? null : lit;
}