'use strict';

var _extends = Object.assign || function (target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i]; for (var key in source) { if (Object.prototype.hasOwnProperty.call(source, key)) { target[key] = source[key]; } } } return target; };

var _require = require('./helpers'),
    getReplacement = _require.getReplacement,
    replace = _require.replace,
    resolveModuleContents = _require.resolveModuleContents,
    isCodegenComment = _require.isCodegenComment,
    isPropertyCall = _require.isPropertyCall,
    requireFromString = _require.requireFromString;

module.exports = getReplacers;

function getReplacers(babel) {

  function asImportCall(path, fileOpts) {
    var _path$get = path.get('arguments'),
        source = _path$get[0],
        args = _path$get.slice(1);

    var _resolveModuleContent2 = resolveModuleContents({
      filename: fileOpts.filename,
      module: source.node.value
    }),
        code = _resolveModuleContent2.code,
        resolvedPath = _resolveModuleContent2.resolvedPath;

    var argValues = args.map(function (a) {
      var result = a.evaluate();
      if (!result.confident) {
        throw new Error('codegen cannot determine the value of an argument in codegen.require');
      }
      return result.value;
    });
    replace({
      path,
      code,
      fileOpts: _extends({}, fileOpts, {
        filename: resolvedPath
      }),
      args: argValues
    }, babel);
  }

  function asTag(path, fileOpts) {
    var code = path.get('quasi').evaluate().value;
    if (!code) {
      throw path.buildCodeFrameError('Unable to determine the value of your codegen string', Error);
    }
    replace({ path, code, fileOpts }, babel);
  }

  function asFunction(path, fileOpts) {
    var argumentsPaths = path.get('arguments');
    var code = argumentsPaths[0].evaluate().value;
    replace({
      path: argumentsPaths[0].parentPath,
      code,
      fileOpts
    }, babel);
  }

  function asJSX(path, fileOpts) {
    var children = path.get('children');
    var code = children[0].node.expression.value;
    if (children[0].node.expression.type === 'TemplateLiteral') {
      code = children[0].get('expression').evaluate().value;
    }
    replace({
      path: children[0].parentPath,
      code,
      fileOpts
    }, babel);
  }

  return {
    asTag,
    asJSX,
    asFunction,
    asProgram: function (path, fileOpts) {
      var _babel$transformFromA = babel.transformFromAst(path.node, {
        filename: fileOpts.filename,
        plugins: fileOpts.plugins,
        presets: fileOpts.presets
      }),
          code = _babel$transformFromA.code;

      var replacement = getReplacement({ code, fileOpts }, babel);
      path.node.body = Array.isArray(replacement) ? replacement : [replacement];
    },
    asImportCall,
    asImportDeclaration: function (path, fileOpts) {
      var codegenComment = path.node.source.leadingComments.find(isCodegenComment).value.trim();

      var _resolveModuleContent = resolveModuleContents({
        filename: fileOpts.filename,
        module: path.node.source.value
      }),
          code = _resolveModuleContent.code,
          resolvedPath = _resolveModuleContent.resolvedPath;

      var args = void 0;
      if (codegenComment !== 'codegen') {
        args = requireFromString(`module.exports = [${codegenComment.replace(/codegen\((.*)\)/, '$1').trim()}]`, fileOpts.filename);
      }
      replace({
        path,
        code,
        fileOpts: _extends({}, fileOpts, {
          filename: resolvedPath
        }),
        args
      }, babel);
    },
    asIdentifier: function (path, fileOpts) {
      var targetPath = path.parentPath;
      switch (targetPath.type) {
        case 'TaggedTemplateExpression':
          {
            return asTag(targetPath, fileOpts);
          }
        case 'CallExpression':
          {
            var isCallee = targetPath.get('callee') === path;
            if (isCallee) {
              return asFunction(targetPath, fileOpts);
            } else {
              return false;
            }
          }
        case 'JSXOpeningElement':
          {
            var jsxElement = targetPath.parentPath;
            return asJSX(jsxElement, fileOpts);
          }
        case 'JSXClosingElement':
          {
            // ignore the closing element
            // but don't mark as unhandled (return false)
            // we already handled the opening element
            return true;
          }
        case 'MemberExpression':
          {
            var callPath = targetPath.parentPath;
            var isRequireCall = isPropertyCall(callPath, 'require');
            if (isRequireCall) {
              return asImportCall(callPath, fileOpts);
            } else {
              return false;
            }
          }
        default:
          {
            return false;
          }
      }
    }
  };
}

/*
eslint
  complexity: ["error", 8]
*/