'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.semanticTrace = semanticTrace;
var arrayPrototypeEnables = {
  map: 1, // not 0 based as we will do null check against this
  every: 1,
  filter: 1,
  find: 1,
  findIndex: 1,
  forEach: 1,
  reduce: 2, // so we dont touch the accumulator - may be issue in future
  reduceRight: 2,
  some: 1
};
var arrayPrototypeIgnores = ['length', 'copyWithin', 'fill', 'pop', // TODO: may want to revisit
'push', 'reverse', 'shift', // TODO: may want to revisit
'unshift', 'sort', // TODO: may want to revisit
'splice', 'concat', 'includes', 'indexOf', 'join', 'lastIndexOf', 'slice', 'toSource', // WARNING PROBABLY DONT USE
'toString', 'toLocaleString', 'entries', // TODO: may want to revisit
'every', // ENABLED
'filter', // ENABLED
'find', // ENABLED
'findIndex', // ENABLED
'forEach', // ENABLED
'keys', 'map', // ENABLED
'reduce', // ENABLED
'reduceRight', // ENABLED
'some', // ENABLED
'values'];

function semanticTrace(referencePath, origin, semanticVisitor) {
  var semanticPath = arguments.length > 3 && arguments[3] !== undefined ? arguments[3] : [];

  var refs = referencePath.scope.bindings[origin].referencePaths.filter(function (ref) {
    return ref.parent !== referencePath.parent;
  });
  // console.log('==', {origin, refs, semanticVisitor})
  conditionalCall(semanticVisitor, referencePath.type, 'origin', referencePath, semanticPath);
  refs.forEach(function (ref) {
    var _traceRHS = traceRHS(ref, semanticPath, semanticVisitor),
        newRef = _traceRHS[0],
        newSemanticPath = _traceRHS[1],
        hasHitArrayMethod = _traceRHS[2];

    if (!hasHitArrayMethod) traceLHS(newRef, newSemanticPath, semanticVisitor);
  });
  conditionalCall(semanticVisitor, 'default', semanticPath);
}

function traceLHS(ref, semanticPath, semanticVisitor) {
  var ptr = ref;
  var newSemanticPath = [].concat(semanticPath);
  if (ptr.parentPath.type === 'VariableDeclarator') {
    var LHS = ptr.parentPath.get('id');
    parseObjectPattern(LHS, newSemanticPath);
  }

  // hoisted up
  function parseObjectPattern(ref, semanticPath) {
    /* --- we do conCall but the semanticVisitor should ideally perform no actions apart from rename  */
    conditionalCall(semanticVisitor, ref.type, 'LHS', ref, semanticPath);
    /* --- we do conCall but should have no actions performed on it  */
    if (ref.type === 'ObjectPattern') {
      var properties = ref.get('properties');
      properties.forEach(function (property) {
        var newSemanticPath = [].concat(semanticPath);
        var key = property.get('key');
        newSemanticPath.push([key.node.name, property]);
        var value = property.get('value');
        parseObjectPattern(value, newSemanticPath);
      });
    } else if (ref.type === 'Identifier') {
      var idname = ref.get('name');
      semanticTrace(ref, idname.node, semanticVisitor, semanticPath);
    }
  }
}

function traceRHS(ref, semanticPath, semanticVisitor) {
  var ptr = ref;
  var newSemanticPath = [].concat(semanticPath);
  var hasHitArrayMethod = false;

  while (isValidRHSParent(ptr)) {
    ptr = ptr.parentPath;
    if (ptr.get('property').type !== 'NumericLiteral')
      // skip number acccess
      workOnRHSParent(ptr, newSemanticPath, semanticVisitor);
  }
  return [ptr, newSemanticPath, hasHitArrayMethod];

  // hoisted up
  function isValidRHSParent(ptr) {
    var baseLayer = ['Member', 'Call'].map(function (x) {
      return `${x}Expression`;
    }).includes(ptr.parentPath.type);
    var validGrandParent = ptr.parentPath.parentPath.type !== 'ExpressionStatement';
    return baseLayer && validGrandParent && !hasHitArrayMethod;
  }

  // hoisted up
  function workOnRHSParent(ptr, newSemanticPath, semanticVisitor) {
    if (ptr.type === 'MemberExpression') {
      var newPath = ptr.node.property.name;
      if (arrayPrototypeIgnores.includes(newPath)) {
        hasHitArrayMethod = true;
        if (arrayPrototypeEnables[newPath] && isValidArrayPrototypeInternal(ptr)) {
          var internalFunctionIndex = arrayPrototypeEnables[newPath]; // get our hardcoded index of supported array methods
          var internalFunction = ptr.parentPath.get('arguments')[0]; // arrow fn
          var paramRef = internalFunction.get('params')[internalFunctionIndex - 1]; // 1-indexed param just to make it null checkable
          var newblade = paramRef.get('name').node; // new blade sits inside the internal function
          if (!newblade && paramRef.type === 'ObjectPattern') {
            // destructuring!
            // *******this is new*******
            parseObjectPattern(paramRef, newSemanticPath);
            // hoisted up
            function parseObjectPattern(ref, semanticPath) {
              /* --- we do conCall but the semanticVisitor should ideally perform no actions apart from rename  */
              conditionalCall(semanticVisitor, ref.type, 'LHS', ref, semanticPath);
              /* --- we do conCall but should have no actions performed on it  */
              if (ref.type === 'ObjectPattern') {
                var properties = ref.get('properties');
                properties.forEach(function (property) {
                  var newSemanticPath = [].concat(semanticPath);
                  var key = property.get('key');
                  newSemanticPath.push([key.node.name, property]);
                  var value = property.get('value');
                  parseObjectPattern(value, newSemanticPath);
                });
              } else if (ref.type === 'Identifier') {
                var idname = ref.get('name');
                semanticTrace(ref, idname.node, semanticVisitor, semanticPath);
              }
            }
            // *******this is new*******
          } else {
            // kick off the traversal inside the internal function
            semanticTrace(paramRef, newblade, semanticVisitor, newSemanticPath);
          }
        }
      } else {
        newSemanticPath.push([newPath, ptr]);
        conditionalCall(semanticVisitor, ptr.type, 'RHS', ptr, newSemanticPath);
      }
    }

    // will be hoisting up
    function isValidArrayPrototypeInternal(ptr) {
      var isValidParent = ptr.parentPath.type === 'CallExpression';
      var isArrowChild = ['ArrowFunctionExpression', 'FunctionExpression'].includes(ptr.parentPath.get('arguments')[0].type);
      return isValidParent && isArrowChild;
    }
  }
}

function conditionalCall(visitor, key) {
  for (var _len = arguments.length, args = Array(_len > 2 ? _len - 2 : 0), _key = 2; _key < _len; _key++) {
    args[_key - 2] = arguments[_key];
  }

  if (visitor[key]) visitor[key].apply(visitor, args);
}